/********************************************************************/
/**                                                                **/
/**               INTEGER MULTIPLICATION (KARATSUBA)               **/
/**                                                                **/
/********************************************************************/

#include "genpari.h"

typedef unsigned long uLong;

/* an integer is an array a[0], ... a[la-1] where la=a[0]>>16 is the number
   of 32-bit words, and  a is a pointer. a[0]&HIM is the true length na,
   a[0]&LOM is the effective length. Then a[1] is the LEAST significant word,
   then a[2], etc... up to a[na-1]. a[na-1] is guaranteed to be non-zero.
   The number 0 is coded [la|1], where la>=1 can be any length. We use GEN
   to denote a pointer on unsigned long, but is NOT a PARI GEN.*/

#define checkmem(z) if (z<bot) err(pile); avma=z;

#define new_chunk(x) (((GEN) avma) - x)

#define LOM 0xffff
#define HIM 0xffff0000

INLINE GEN
maget(long x)
{
  const GEN z = new_chunk(x);

  checkmemory((ulong)z);
  z[0] = x|(x<<16);
  return z;
}

void
maadd(GEN a, GEN b, GEN z)
{
  long na,nb;
  GEN p1;

  if (a[0]&LOM<b[0]&LOM) { p1=a; a=b; b=p1; }
  na=a[0]&LOM; nb=b[0]&LOM;
/* now na>=nb */
  if (nb==1)
  {
    for (i=1; i<na; i++) z[i]=a[i];
    z[0]=(z[0]&HIM)|na;
  }
  else
  {
    z[1]=addllfast(a[1],b[1]);
    for (i=2; i<nb; i++) z[i]=addllxfast(a[i],b[i]);
    for ( ; (i<na)&&overf; i++) z[i]=addllxfast0(a[i]);
    if (overf) z[na]=1;
    else for ( ; i<na; i++) z[i]=a[i];
    z[0]=(z[0]&HIM)|(na+overf);
  }
}

void
masub(GEN a, GEN b, GEN z)
{
/* assume a>=b */
  long na,nb;
  GEN p1;

  na=a[0]&LOM; nb=b[0]&LOM;
  DEBUG_ASSERT(na>=nb);
/* now na>=nb */
  if (nb==1)
  {
    for (i=1; i<na; i++) z[i]=a[i];
    z[0]=(z[0]&HIM)|na;
  }
  else
  {
    z[1]=subllfast(a[1],b[1]);
    for (i=2; i<nb; i++) z[i]=subllxfast(a[i],b[i]);
    for ( ; (i<na)&&overf; i++) z[i]=subllxfast0(a[i]);
    DEBUG_ASSERT(overf==0);
    if (i<na)
    {
      for ( ; i<na; i++) z[i]=a[i];
      z[0]=(z[0]&HIM)|na;
    }
    else
    {
      for (i=na-1; (i>=2)&&(!z[i]); i--);
      z[0]=(z[0]&HIM)|(i+1);
    }
  }
}

int
macmp(GEN a, GEN b)
{
/* sign(a-b) */
  long na, nb, dif;
  
  na=a[0]&LOM; nb=b[0]&LOM;  
  if (na>nb) return 1;
  if (na<nb) return -1;
/* here na==nb */
  for (i=na-1; (i>=1)&&(a[i]==b[i]); i--);
  if (!i) return 0;
  return (a[i]>b[i]) ? 1: -1;
}


    


